# Copyright 2023 Inductor, Inc.
"""Functions for running quality measures."""

from typing import Any

import openai
import tiktoken

from inductor import wire_model

class Inputs(dict):
    """A dictionary that allows access to its keys as attributes."""

    def __getattribute__(self, key: str) -> Any:
        """Return the value corresponding to key in this dict.
        
        Args:
            key: The key to access.

        Raises:
            AttributeError: If the key does not exist in this dict.
        """
        try:
            return self[key]
        except KeyError as exc:
            raise AttributeError(key) from exc


def openai_llm_quality_measure(
    output: str,
    inputs: Inputs,
    test_case: wire_model.TestCase,
    quality_measure: wire_model.QualityMeasure,
    model: str,
    prompt: str
) -> str:
    """Evaluate an LLM-powered quality measure using the OpenAI API.
    
    Args:
        output: The output of the execution of a LLM program.
        inputs: The inputs to the LLM program execution that produced output.
        test_case: The test case that served as input to the LLM program
            execution that produced output.
        quality_measure: The quality measure being evaluated.
        model: The model to use in the LLM powered quality measure.
        prompt: The prompt for the LLM powered quality measure.

    Returns:
        The result of evaluating the quality measure.
    """
    client = openai.OpenAI()
    accepted_outputs = []
    evaluation_type_prompt = ""
    if quality_measure.evaluation_type == "BINARY":
        evaluation_type_prompt = (
            "Output must be *ONLY* one of Y (for YES) or N (for NO).")
        accepted_outputs = ["Y", "N"]
    elif quality_measure.evaluation_type == "RATING_INT":
        evaluation_type_prompt = (
            "Output must be *ONLY* an integer from 1 through 5.")
        accepted_outputs = ["1", "2", "3", "4", "5"]

    encoding = tiktoken.encoding_for_model(model)
    max_tokens = 1
    logit_bias = {}
    for accepted in accepted_outputs:
        tokens = encoding.encode(accepted)
        max_tokens = max(max_tokens, len(tokens))
        logit_bias.update({token: 100 for token in tokens})

    execution = {"output": output, "inputs": inputs, "test_case": test_case}
    chat_completion = client.chat.completions.create(
        messages=[
            {"role": "system", "content": prompt.format_map(execution)},
            {"role": "system", "content": evaluation_type_prompt}
        ],
        model=model,
        max_tokens=max_tokens,
        logit_bias=logit_bias)
    return chat_completion.choices[0].message.content



# Copyright 2023 Inductor, Inc.
"""Inductor client library."""

import contextlib
import contextvars
import copy
import datetime
import functools
import inspect
import io
import sys
from typing import Any, Callable, Dict, Iterator, List, Optional, TextIO, Tuple, TypeVar, Union

from inductor import auth_session, backend_client, config, wire_model
from inductor.cli import cli, data_model
from inductor.quality_measures import Inputs
from inductor.wire_model import TestCase, QualityMeasure, HparamSpec


# Test suite components are classes that can be added to a test suite,
# i.e., test cases, quality measures, and hyperparameter specifications.
_TestSuiteComponent = Union[TestCase, QualityMeasure, HparamSpec]


class TestSuite:
    """Test suite.
    
    A collection of test cases, quality measures, and hyperparameter
    specifications that can be run together.
    
    Attributes:
        id_or_name: ID or name of the test suite.
        llm_program: LLM program to test. Either a Python object or a string
            containing the fully qualified name of the Python object. The
            Python object can be either a Python function or LangChain chain.
            If a string is passed, it must be in the format:
            `<fully qualified module name>:<fully qualified object name>`.
        test_cases: List of test cases to run.
        quality_measures: List of quality measures to compute.
        hparam_specs: List of hyperparameter specifications.
    """
    def __init__(
        self,
        id_or_name: Union[int, str],
        llm_program: Union[Callable, str]):
        """Create a test suite.
        
        Args:
            id_or_name: ID or name of the test suite.
            llm_program: LLM program to test. Either a Python object or a
                string containing the fully qualified name of the Python
                object. The Python object can be either a Python function or
                LangChain chain. If a string is passed, it must be in the
                format:
                `<fully qualified module name>:<fully qualified object name>`.
        """
        self.id_or_name = id_or_name
        self.llm_program = data_model.LazyCallable(llm_program)

        self.test_cases = []
        self.quality_measures = []
        self.hparam_specs = []


    def add(self, *args: Union[_TestSuiteComponent, List[_TestSuiteComponent]]):
        """Add test cases, quality measures, or hyperparameter specifications.
        
        Args:
            *args: One or more test cases, quality measures, or hyperparameter
                specifications to add to the test suite. If a list is passed,
                each item in the list is added to the test suite.
        """
        for arg in args:
            if isinstance(arg, TestCase):
                self.test_cases.append(arg)
            elif isinstance(arg, QualityMeasure):
                self.quality_measures.append(arg)
            elif isinstance(arg, HparamSpec):
                self.hparam_specs.append(arg)
            elif isinstance(arg, list):
                for item in arg:
                    self.add(item)
            else:
                raise TypeError(
                    "Invalid type. Expected TestCase, QualityMeasure, "
                    f"or HparamSpec, but got {type(arg)}.")

    def run(self, replicas: int = 1, parallelize: int = 1):
        """Run the test suite.

        Args:
            replicas: Number of replicated executions to perform for each
                (test case, unique set of hyperparameter values) pair.
            parallelize: Number of LLM program executions to run in parallel.
        """
        run_request = wire_model.CreateTestSuiteRunRequest(
            test_suite_id_or_name=self.id_or_name,
            test_cases=self.test_cases,
            quality_measures=self.quality_measures,
            hparam_specs=self.hparam_specs,
            llm_program_details=self.llm_program.get_details(),
            replicas=replicas,
            parallelize=parallelize,
        )
        cli.execute_test_suite_run(
            test_suite_run=run_request,
            auth_access_token=auth_session.get_auth_session().access_token,
        )


# The following module-private variables are used by the functions in
# the rest of this module to transmit information to and from LLM program
# executions.
# Whether the logger decorator (inductor.logger) is enabled. This is set to
# False when running tests to prevent the logger from sending duplicate data to
# the backend, in the case that the LLM program being tested uses the logger
# decorator.
_logger_decorator_enabled = True
# Context variable used to store the logged values for the current LLM program
# execution. This is a context variable instead of a global variable so that
# the logger will work correctly when running mutliple threads that each use
# the logger decorator. However, an exception will be raised if the logger
# decorated function itself uses multiple threads that each call inductor.log.
_logged_values = contextvars.ContextVar("logged_values", default=None)
# Dictionary of hyperparameter values for the current LLM program execution.
_hparams = {}
# Context variable used to store whether the current LLM program execution is
# the primary execution.
_primary_execution = contextvars.ContextVar("active_execution", default=True)


def hparam(name: str, default_value: Any) -> Any:
    """Return the value of the hyperparameter having the given name.

    Args:
        name: Name of hyperparameter value to be returned.
        default_value: Value that will be returned if a value has not
            been specified for the given name.
    """
    return _hparams.get(name, default_value)


def _log(
    value: Any, *, after_complete: bool, description: Optional[str] = None):
    """Log a value and associate it with the current LLM program execution.

    Args:
        value: The value to be logged.
        after_complete: Whether the value was logged after the LLM
            program execution completed.
        description: An optional human-readable description of the logged
            value.
    
    Raises:
        RuntimeError: If the LLM program execution was not initiated via the
            Inductor CLI, and the LLM program is not decorated with
            @inductor.logger.
    """
    logged_values = _logged_values.get()
    if logged_values is None:
        # We can not distinguish between the below two cases described in the
        # exception message, so we raise the same exception in both cases.
        raise RuntimeError(
            "Cannot call inductor.log outside of a function decorated with "
            "@inductor.logger, unless you are running `inductor test`. "
            "Also note that invoking inductor.log from a thread different "
            "from the one that initialized the logger (via the decorator or "
            "the CLI tool) is currently unsupported. If you require support "
            "for this, please contact Inductor support to submit a feature "
            "request.")
    logged_values.append(
        wire_model.LoggedValue(
            value=copy.deepcopy(value),
            description=description,
            after_complete=after_complete))


def log(value: Any, *, name: Optional[str] = None):
    """Log a value and associate it with the current LLM program execution.

    Args:
        value: The value to be logged.
        name: An optional human-readable name for the logged value.
    
    Raises:
        RuntimeError: If the LLM program execution was not initiated via the
            Inductor CLI, and the LLM program is not decorated with
            @inductor.logger.
    """
    _log(value, description=name, after_complete=False)


@contextlib.contextmanager
def _configure_for_test(hparams: Dict[str, Any]):
    """Configure the Inductor library for a test suite run.
    
    Disable the inductor.logger decorator by setting
    `inductor._logger_decorator_enabled` to False and set the inductor._hparams
    to the given hyperparameters. On exit, restore the original value of
    `inductor._logger_decorator_enabled` and set `inductor._hparams` to an
    empty dictionary.

    Args:
        hparams: A dictionary mapping hyperparameter names to values.
    """
    global _hparams
    global _logger_decorator_enabled
    orig_logger_decorator_enabled = _logger_decorator_enabled
    try:
        _hparams = hparams
        _logger_decorator_enabled = False
        yield
    finally:
        _hparams = {}
        _logger_decorator_enabled = orig_logger_decorator_enabled


@contextlib.contextmanager
def _capture_logged_values():
    """Capture values logged via log() calls.
    
    If logging has not already been initialized, initialize logging by setting
    the logged values context variable (`_logged_values`) to an empty list,
    and, on exit, set `_logged_values` to `None`.
    If logging has already been initialized, do nothing.
    In either case, yield the list of logged values.

    The purpose of this context manager is to manage the state of the
    logged values context variable, which should only be initialized
    once per LLM program execution.

    Yields:
        The list of logged values.
    """
    logged_values = _logged_values.get()
    initializing_logged_values = logged_values is None
    try:
        if initializing_logged_values:
            _logged_values.set([])
        yield _logged_values.get()
    finally:
        if initializing_logged_values:
            _logged_values.set(None)


@contextlib.contextmanager
def _capture_stdout_stderr(
    suppress: bool = False) -> Tuple[io.StringIO, io.StringIO]:
    """Capture stdout and stderr.
    
    On exit, restore the original stdout and stderr and close the yielded
    StringIO buffers (i.e., the yielded buffers' contents will be discarded
    when context manager exits).
    
    Args:
        suppress: Whether to suppress stdout and stderr. If True, the
            contents of stdout and stderr will be suppressed after being
            captured. If False, stdout and stderr will behave as normal,
            but their contents will still be captured.

    Yields:
        A tuple of streams used to capture stdout and stderr.
    """
    class Tee(io.StringIO):
        """A StringIO buffer that optionally writes to a file in addition to
        capturing the written string."""
        def __init__(self, file: Optional[TextIO]):
            """Override the constructor to store the file to which to write."""
            self.file = file
            super().__init__()

        def write(self, s: str):
            """Override the write method to write to the file (as merited)
            in addition to capturing the written string."""
            if self.file is not None:
                self.file.write(s)
            return super().write(s)

    stdout_capture = Tee(
        sys.stdout if not suppress else None)
    stderr_capture = Tee(
        sys.stderr if not suppress else None)

    # Save the original stdout and stderr.
    original_stdout = sys.stdout
    original_stderr = sys.stderr
    # Redirect stdout and stderr to the Tee objects.
    sys.stdout = stdout_capture
    sys.stderr = stderr_capture
    try:
        yield (stdout_capture, stderr_capture)
    finally:
        # Restore the original stdout and stderr.
        sys.stdout = original_stdout
        sys.stderr = original_stderr
        # Close the StringIO buffers.
        stdout_capture.close()
        stderr_capture.close()


@contextlib.contextmanager
def _manage_executions():
    """Manage the state of the primary execution context variable.

    Manage the state of the primary execution context variable
    (_primary_execution). If the variable is initially True, it is set to
    False and True is yielded. If the variable is initially False, False is
    yielded. On exit, the variable is restored to its original value.

    The purpose of this context manager is to allow the logger decorator to
    determine whether it is the primary (top-level) execution. This is
    necessary because the logger decorator should only send data to the
    backend if it is the primary execution. For example, when the logger
    decorator decorates a function that is called by another function also
    decorated with the logger decorator, the logger decorator should not send
    data to the backend during the inner function call.

    Yields:
        True if the primary execution context variable was True, False
        otherwise.
    """
    primary_execution = _primary_execution.get()
    if primary_execution:
        _primary_execution.set(False)
    try:
        yield primary_execution
    finally:
        _primary_execution.set(primary_execution)


def _log_completed_execution(
    *,
    output: Any,
    primary_execution: bool,
    llm_program: data_model.LazyCallable,
    input_args: Dict[str, Any],
    error: Optional[str] = None,
    stdout: Optional[str] = None,
    stderr: Optional[str] = None,
    logged_values: Optional[List[wire_model.LoggedValue]] = None,
    started_at: datetime.datetime,
    auth_access_token: str,
    ) -> Any:
    """Log the completion of an LLM program execution.

    If the LLM program execution is the primary execution, send the execution
    data to the backend. Otherwise, call the Inductor client's `log()`
    function to log this execution as part of the current overarching
    primary execution.

    Args:
        output: Result of the LLM program execution.
        primary_execution: Whether the LLM program execution is the primary
            execution.
        llm_program: LLM program that was executed.
        input_args: Input arguments to the LLM program.
        error: Error message, if any, that occurred during the LLM program
            execution.
        stdout: Captured stdout from the LLM program execution.
        stderr: Captured stderr from the LLM program execution.
        logged_values: Values logged during the LLM program execution.
        started_at: Time at which the LLM program execution started.
        auth_access_token: Access token used to authenticate the request to
            the backend.
    """
    ended_at = datetime.datetime.now(datetime.timezone.utc)

    if primary_execution:
        backend_client.log_llm_program_execution_request(
            wire_model.LogLlmProgramExecutionRequest(
                program_details=llm_program.get_details(),
                execution_details=wire_model.ExecutionDetails(
                    mode="DEPLOYED",
                    inputs=input_args,
                    hparams=_hparams or None,
                    output=output,
                    error=error,
                    stdout=stdout,
                    stderr=stderr,
                    logged_values=logged_values or None,
                    execution_time_secs=(
                        ended_at - started_at).total_seconds(),
                    started_at=started_at,
                    ended_at=ended_at,)),
            auth_access_token)

    else:
        log(
            {
                "llm_program":
                llm_program.fully_qualified_name,
                "inputs": input_args,
                "output": output
            },
            name="Nested LLM program execution")


def logger(func: Callable) -> Callable:
    """Log the inputs, outputs, and inductor.log calls of func.

    Use `logger` as a decorator to automatically log the arguments and return
    value of, as well as calls to inductor.log within, the decorated function.
    For example:
        @inductor.logger
        def hello_world(name: str) -> str:
            inductor.log(len(name), description="name length")
            return f"Hello {name}!"

    Args:
        func: The decorated function.
    
    Returns:
        Wrapped function.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        if _logger_decorator_enabled:
            with (
                _capture_logged_values() as logged_values,
                _manage_executions() as primary_execution,
                # TODO: We don't need to capture stdout and stderr if we are
                # not in the primary execution. However since the stdout and
                # stderr are not suppressed, the user will not notice the
                # difference.
                _capture_stdout_stderr(suppress=False) as (stdout, stderr)
            ):
                auth_access_token = auth_session.get_auth_session().access_token

                llm_program = data_model.LazyCallable(func)

                # Get input arguments using the function's signature.
                signature = inspect.signature(func)
                bound_arguments = signature.bind(*args, **kwargs)
                bound_arguments.apply_defaults()
                input_args = copy.deepcopy(bound_arguments.arguments)

                started_at = datetime.datetime.now(datetime.timezone.utc)
                result = None
                error = None
                try:
                    result = func(*args, **kwargs)

                    # If the result is an iterator, we cannot make the
                    # assumption that the iterator can be re-used.
                    # Therefore we return a wrapper for the iterator that
                    # captures returned values and logs them when the
                    # iterator is exhausted.
                    if isinstance(result, Iterator):
                        _T_IteratorWrapper = TypeVar(
                            "_T_IteratorWrapper", bound="_IteratorWrapper")  # pylint: disable=invalid-name
                        class _IteratorWrapper:
                            """Iterator wrapper that captures yielded values.

                            When the iterator is exhausted, the wrapper logs
                            the completion of the LLM program execution to
                            Inductor.
                            """
                            def __init__(
                                self,
                                iterator: Iterator,
                                stdout: Optional[str] = None,
                                stderr: Optional[str] = None):
                                """Create an _IteratorWrapper.
                                
                                Args:
                                    iterator: The iterator to wrap.
                                    stdout: Captured stdout from the LLM
                                        program execution.
                                    stderr: Captured stderr from the LLM
                                        program execution.
                                """
                                self._iterator = iterator
                                self._completed_values = []
                                self._stdout = stdout
                                self._stderr = stderr

                            def __iter__(self) -> _T_IteratorWrapper:
                                return self

                            def __next__(self) -> Any:
                                """Get the next value from the iterator.

                                Add the returned value to the list of completed
                                values. If the iterator is exhausted, log the
                                completion of the LLM program execution.

                                Returns:
                                    The next value from the iterator.
                                """
                                try:
                                    value = next(self._iterator)
                                    self._completed_values.append(
                                        copy.deepcopy(value))
                                except StopIteration as stop_signal:
                                    if all(
                                        isinstance(value, str)
                                        for value in self._completed_values
                                    ):
                                        self._completed_values = "".join(
                                            self._completed_values)
                                    _log_completed_execution(
                                        output=self._completed_values,
                                        primary_execution=primary_execution,
                                        llm_program=llm_program,
                                        input_args=input_args,
                                        started_at=started_at,
                                        stdout=self._stdout,
                                        stderr=self._stderr,
                                        logged_values=logged_values,
                                        auth_access_token=auth_access_token,
                                    )
                                    raise stop_signal
                                return value

                            def __getattr__(self, name: str) -> Any:
                                """Forward unknown attributes to the iterator.
                                
                                This allows the iterator wrapper to be used
                                as if it were the underlying iterator.

                                Args:
                                    name: Name of attribute to get.
                                
                                Returns:
                                    The requested attribute.
                                """
                                return getattr(self._iterator, name)

                        return _IteratorWrapper(
                            result,
                            stdout.getvalue(),
                            stderr.getvalue())

                except Exception as e:  # pylint: disable=broad-except
                    error = str(e)

                _log_completed_execution(
                    output=copy.deepcopy(result),
                    primary_execution=primary_execution,
                    llm_program=llm_program,
                    input_args=input_args,
                    started_at=started_at,
                    error=error,
                    stdout=stdout.getvalue(),
                    stderr=stderr.getvalue(),
                    logged_values=logged_values,
                    auth_access_token=auth_access_token,
                )
                return result
        else:
            return func(*args, **kwargs)
    return wrapper



# Copyright 2023 Inductor, Inc.
"""Functionality for executing a test suite run."""

import datetime
import itertools
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import loky

import inductor
from inductor.cli import data_model
from inductor import backend_client, quality_measures, wire_model


def _execute_quality_measure_llm(
    llm_program_output: Any,
    quality_measure: wire_model.QualityMeasure,
    test_case: wire_model.TestCase,
) -> Union[str, bool, int]:
    """Execute a LLM powered quality measure on the LLM program output.

    Args:
        llm_program_output: Output of the LLM program.
        quality_measure: Quality measure.  Must have evaluator field equal
            to "LLM".
        test_case: Test case on which the LLM program was executed to
            produce llm_program_output.
    
    Raises:
        ValueError: If quality measure evaluator is not "LLM".

    Returns:
        Output of the quality measure.
    """
    if quality_measure.evaluator != "LLM":
        raise ValueError("Quality measure evaluator must be 'LLM'.")

    evaluation_type = quality_measure.evaluation_type
    if isinstance(quality_measure.spec, str):
        callable_object = data_model.LazyCallable(quality_measure.spec)
        args = [
            llm_program_output,
            quality_measures.Inputs(test_case.inputs),
            test_case,
            quality_measure,
        ]
        inputs_signature = callable_object.inputs_signature
        quality_measure_output = callable_object(*args[:len(inputs_signature)])

    elif isinstance(quality_measure.spec, dict):
        model = quality_measure.spec.get("model", "gpt-3.5-turbo")
        prompt = quality_measure.spec.get("prompt")
        if prompt is None:
            raise ValueError(
                "Quality measure spec must contain a 'prompt' key.")
        quality_measure_output = quality_measures.openai_llm_quality_measure(
            llm_program_output,
            model=model,
            prompt=prompt,
            inputs=quality_measures.Inputs(test_case.inputs),
            quality_measure=quality_measure,
            test_case=test_case)

    if isinstance(quality_measure_output, str):
        if evaluation_type == "BINARY":
            if quality_measure_output.strip().upper() in ("YES", "Y"):
                quality_measure_output = True
            elif quality_measure_output.strip().upper() in ("NO", "N"):
                quality_measure_output = False

        elif evaluation_type == "RATING_INT":
            quality_measure_output = int(quality_measure_output)

    return quality_measure_output


def _execute_test_case(
    test_suite_run_id: int,
    llm_program_fully_qualified_name: str,
    test_case: wire_model.TestCase,
    test_case_id: int,
    test_case_replica_index: int,
    quality_measures: List[wire_model.QualityMeasure],  # pylint: disable=redefined-outer-name
    quality_measure_ids: List[int],
    hparams: Dict[str, Any],
    auth_access_token: str,
) -> Tuple[
    wire_model.LogTestCaseExecutionRequest,
    List[Dict[str, Union[
        int, str, wire_model.QualityMeasureExecutionDetails]]]]:
    """Run a test case and evaluate its quality measures.

    Sends the output of the test case and the outputs of the quality measures
    to the backend server.

    Args:
        test_suite_run_id: ID of test suite run.
        llm_program_fully_qualified_name: Fully qualified name of LLM program.
        test_case: Test case.
        test_case_id: ID of test case.
        test_case_replica_index: Index of test case replica.
        quality_measures: List of quality measures.
        quality_measure_ids: IDs of quality measures.
        hparams: Mapping from hyperparameter names to values.
        auth_access_token: Auth0 access token.
    
    Returns:
        A tuple of the LogTestCaseExecutionRequest and a list of invalid
            quality measures if any. Each invalid quality measure is a
            Dictionary with the following keys:
                "id": ID of quality measure.
                "name": Name of quality measure.
                "execution_details": wire_model.QualityMeasureExecutionDetails
                    object.
    """
    started_at = datetime.datetime.now(datetime.timezone.utc)

    llm_program_stdout = None
    llm_program_stderr = None
    llm_program_error = None
    llm_program_output = None
    with (
        inductor._capture_stdout_stderr(suppress=True) as (stdout, stderr),  # pylint: disable=protected-access
        inductor._capture_logged_values() as logged_values,  # pylint: disable=protected-access
        inductor._configure_for_test(hparams)  # pylint: disable=protected-access
    ):
        try:
            # Run the LLM program.
            llm_program = data_model.LazyCallable(
                llm_program_fully_qualified_name)
            llm_program_output = llm_program(**test_case.inputs)
            if isinstance(llm_program_output, Iterator):
                llm_program_output = list(llm_program_output)
                if all(isinstance(value, str) for value in llm_program_output):
                    llm_program_output = "".join(llm_program_output)

        except Exception as error:  # pylint: disable=broad-except
            llm_program_error = str(error)

        llm_program_stdout = stdout.getvalue()
        llm_program_stderr = stderr.getvalue()

    ended_at = datetime.datetime.now(datetime.timezone.utc)

    # If the LLM program completed without error, run the executable quality
    # measures to generate a list of direct evaluations.
    direct_evaluations = []
    # For any executable quality measures that raises an error, or returns a
    # non-integer, non-boolean output, add it to the list of invalid quality
    # measures. TODO: Log quality measure errors on the backend.
    invalid_quality_measures = []
    if llm_program_error is None:

        # Run the executable quality measures.
        quality_measures_outputs = []
        for quality_measure, quality_measure_id in zip(
            quality_measures, quality_measure_ids):
            if quality_measure.evaluator in ("FUNCTION", "LLM"):
                # TODO: Record quality measure errors, stdout, and stderr on
                # the backend. See wire_model.QualityMeasureExecutionDetails
                # for the additional TODO of how these should be recorded.
                quality_measure_stdout = None
                quality_measure_stderr = None
                quality_measure_error = None
                quality_measure_output = None
                with inductor._capture_stdout_stderr(  # pylint: disable=protected-access
                    suppress=True) as (stdout, stderr):
                    try:
                        if quality_measure.evaluator == "FUNCTION":
                            callable_object = data_model.LazyCallable(
                                quality_measure.spec)
                            quality_measure_output = callable_object(
                                llm_program_output)

                        elif quality_measure.evaluator == "LLM":
                            quality_measure_output = (
                                _execute_quality_measure_llm(
                                    llm_program_output,
                                    quality_measure,
                                    test_case))

                    except Exception as error:  # pylint: disable=broad-except
                        quality_measure_error = str(error)

                    quality_measure_stdout = stdout.getvalue()
                    quality_measure_stderr = stderr.getvalue()

                quality_measure_execution_details = (
                    wire_model.QualityMeasureExecutionDetails(
                        input=llm_program_output,
                        output=quality_measure_output,
                        error=quality_measure_error,
                        stdout=quality_measure_stdout,
                        stderr=quality_measure_stderr,
                    )
                )

                if quality_measure_error is not None:
                    invalid_quality_measures.append({
                        "id": quality_measure_id,
                        "name": quality_measure.name,
                        "execution_details": quality_measure_execution_details,
                    })

                else:
                    quality_measures_outputs.append((
                        quality_measure_id,
                        quality_measure,
                        quality_measure_execution_details,
                    ))


        # Create direct evaluations for the quality measures that are
        # executable and valid. If a quality measure that is executable
        # returns an output type that does not match the quality measure's
        # evaluation type, then add it to the list of invalid quality
        # measures and do not create a direct evaluation for it.
        for (
            quality_measure_id,
            quality_measure,
            quality_measure_execution_details,
        ) in quality_measures_outputs:
            quality_measure_output = quality_measure_execution_details.output
            if quality_measure.evaluation_type == "BINARY" and isinstance(
                quality_measure_output, bool):
                direct_evaluations.append(
                    wire_model.DirectEvaluation(
                        quality_measure_id=quality_measure_id,
                        value_bool=quality_measure_output))
            elif (quality_measure.evaluation_type == "RATING_INT" and
                    isinstance(quality_measure_output, int) and
                    # Required to prevent `bool` from being interpreted as
                    # `int`, since `bool` is a subclass of `int`.
                    not isinstance(quality_measure_output, bool)):
                direct_evaluations.append(
                    wire_model.DirectEvaluation(
                        quality_measure_id=quality_measure_id,
                        value_int=quality_measure_output))
            else:
                expected_output_type = (
                    type(True) if quality_measure.evaluation_type == "BINARY"
                    else type(5))
                quality_measure_execution_details.error = (
                    f"Invalid output type. Expected output type: "
                    f"{expected_output_type}. Actual output type: "
                    f"{type(quality_measure_output)}")
                invalid_quality_measures.append({
                    "id": quality_measure_id,
                    "name": quality_measure.name,
                    "execution_details": quality_measure_execution_details,
                })

    request_object = wire_model.LogTestCaseExecutionRequest(
        test_suite_run_id=test_suite_run_id,
        test_case_id=test_case_id,
        test_case_replica_index=test_case_replica_index,
        execution_details=wire_model.ExecutionDetails(
            mode="CLI",
            inputs=test_case.inputs,
            hparams=hparams or None,
            output=llm_program_output,
            error=llm_program_error,
            stdout=llm_program_stdout,
            stderr=llm_program_stderr,
            execution_time_secs=(ended_at - started_at).total_seconds(),
            started_at=started_at,
            ended_at=ended_at,
            logged_values=logged_values or None,
            direct_evaluations=direct_evaluations or None,
        )
    )

    backend_client.log_test_case_execution(request_object, auth_access_token)
    return request_object, invalid_quality_measures

def get_hparams_combinations(
    hparam_specs: Optional[List[wire_model.HparamSpec]] = None
    ) -> List[Dict[str, Any]]:
    """Get all combinations of hyperparameters.

    Given a list of hyperparameters and their possible values, return a list of
    dictionaries, where each dictionary represents a unique combination of
    hyperparameters.

    For example, if the given hyperparameters are:
    [
        wire_model.HparamSpec(
            hparam_name="a",
            values=[1, 2],
        ),
        wire_model.HparamSpec(
            hparam_name="b",
            values=[3, 4],
        ),
    ]
    then the returned list will be:
    [
        {"a": 1, "b": 3},
        {"a": 1, "b": 4},
        {"a": 2, "b": 3},
        {"a": 2, "b": 4},
    ]

    Args:
        hparam_specs: List of hyperparameters specs, where all hyperparameter
            specs have distinct names.

    Returns:
        A list of dictionaries, where each dictionary represents a unique
            combination of hyperparameters.

    Raises:
        ValueError: If hyperparameter names are not distinct.
    """
    if hparam_specs is None:
        return [{}]

    # Convert list of HparamSpec to dictionary
    hparams_dict = {
        hparam.hparam_name: hparam.values for hparam in hparam_specs}

    # Ensure that all hyperparameter names in hparam_specs are distinct.
    if len(hparams_dict) != len(hparam_specs):
        raise ValueError(
            "Hyperparameter names in hparam_specs must be distinct.")

    keys = list(hparams_dict.keys())
    value_lists = [hparams_dict[key] for key in keys]

    # Generate combinations.
    value_combinations = list(itertools.product(*value_lists))

    # Convert to dictionaries.
    hparam_combinations = []
    for value_combination in value_combinations:
        hparam_combinations.append(dict(zip(keys, value_combination)))

    return hparam_combinations


def execute_test_suite_run(
    test_suite_run: wire_model.CreateTestSuiteRunRequest,
    test_suite_run_metadata: wire_model.CreateTestSuiteRunResponse,
    executor: loky.Executor,
    auth_access_token: str,
) -> List[loky.Future]:
    """Execute a test suite run.

    Args:
        test_suite_run: Test suite run.
        test_suite_run_metadata: Test suite run metadata.
        executor: Executor for running test cases.
        auth_access_token: Auth0 access token.

    Returns:
        List of futures for the test case outputs.
    """
    test_case_futures = []
    for test_case_replica_index in range(test_suite_run.replicas):
        for hparams in get_hparams_combinations(test_suite_run.hparam_specs):
            for test_case, test_case_id in zip(
                test_suite_run.test_cases,
                test_suite_run_metadata.test_case_ids):
                test_case_futures.append(executor.submit(
                    _execute_test_case,
                    test_suite_run_id=
                        test_suite_run_metadata.test_suite_run_id,
                    llm_program_fully_qualified_name=
                        test_suite_run.llm_program_details.fully_qualified_name,
                    test_case=test_case,
                    test_case_id=test_case_id,
                    test_case_replica_index=test_case_replica_index,
                    quality_measures=test_suite_run.quality_measures,
                    quality_measure_ids=
                        test_suite_run_metadata.quality_measure_ids,
                    hparams=hparams,
                    auth_access_token=auth_access_token,
                ))
    return test_case_futures



# Copyright 2023 Inductor, Inc.
"""Collection of Questions / User Prompts for the Inductor CLI.

This module is intended to include questions that require non-trivial
validation or other complex logic. It does not need to include simple
questions.
"""

import pathlib
import re
from typing import Any, Dict, Optional
import inquirer
import rich

from inductor.cli import data_model


def get_new_test_suite_file_path() -> pathlib.Path:
    """Prompt user for a path to a new test suite file.

    Prompt the user to enter a path for a new test suite file. Validate the
    path (see `validate_path` function below). If the path is valid, return
    it. If the path is invalid, continue prompting the user until a valid
    path is entered.

    Returns:
        Path to the test suite file.
    """
    default_path = pathlib.Path("test_suite.yaml")
    i = 1
    while default_path.exists():
        default_path = pathlib.Path(f"test_suite_{i}.yaml")
        i += 1
        if i > 100:
            # If we've tried 100 times, something is wrong.
            # Use the initial default path.
            default_path = pathlib.Path("test_suite.yaml")
            break

    # TODO (https://github.com/inductor-hq/saas/issues/18): Add autocomplete
    # for paths.
    def validate_path(_: Dict[str, Any], current: str) -> bool:
        """Validate a given path.

        A path is valid if it:
        1. Has a .yaml or .yml extension.
        2. Does not already exist.
        3. Can be touched.

        Given a path to a file, validate it. If the path is valid and gets
        created, delete it.

        Args:
            _: Dictionary containing the answers from previous questions
                (not including the current question). This will be an empty
                dictionary because the `test_suite_file_path` question is not
                part of a collection of questions.
            current: The current answer, which is a path to validate.
        
        Returns:
            True if the path is valid, False otherwise.
        
        Raises:
            inquirer.errors.ValidationError: If the path is invalid.
        """
        try:
            path = pathlib.Path(current).resolve()
            if path.suffix not in [".yml", ".yaml"]:
                # NOTE: This will also catch empty paths.
                raise inquirer.errors.ValidationError(
                    "",
                    reason="File requires a .yaml or .yml extension.")

            # Permissions: rw-rw-rw- (0o666)
            path.touch(mode=0o666, exist_ok=False)
            if path.exists():
                path.unlink()
        except FileExistsError:
            raise inquirer.errors.ValidationError(  # pylint: disable=raise-missing-from
                "", reason="File already exists.")
        except PermissionError:
            raise inquirer.errors.ValidationError(  # pylint: disable=raise-missing-from
                "", reason="Missing permissions to create this file.")
        except FileNotFoundError:
            # The path specified by current does not exist.
            raise inquirer.errors.ValidationError(  # pylint: disable=raise-missing-from
                "", reason="Invalid path.")
        return True

    rich.print("\n[[yellow]1/4[/yellow]] Enter the file path for the new test "
               "suite, ensuring it has a .yaml or .yml extension and does "
               "not already exist.")
    test_suite_file_path = inquirer.text(
        message="Test suite file path",
        validate=validate_path,
        default=default_path)
    return pathlib.Path(test_suite_file_path).resolve()


def get_test_suite_name() -> str:
    """Prompt user for a test suite name.

    Prompt the user to enter a test suite name. Validate the name (see
    `validate_test_suite_name` function below). If the name is valid,
    return it. If the name is invalid, continue prompting the user until a
    valid name is entered.

    Returns:
        Test suite name.
    """
    def validate_test_suite_name(_: Dict[str, Any], current: str) -> bool:
        """Validate test suite name.

        A test suite name is valid if it:
        1. Is non-empty.
        2. Includes only alphanumeric characters, underscores, or dashes.

        Args:
            _: Dictionary containing the answers from previous questions
                (not including the current question). This will be an empty
                dictionary because the `test_suite_name` question is not
                part of a collection of questions.
            current: The current answer, which is a test suite name to
                validate.

        Returns:
            True if the test suite name is valid, False otherwise.
        
        Raises:
            inquirer.errors.ValidationError: If the test suite name is
                invalid.
        """
        name = current

        if not re.match(r"^[a-zA-Z0-9_-]+$", name):
            raise inquirer.errors.ValidationError(
                "", reason="Invalid test suite name.")
        return True

    rich.print("\n[[yellow]2/4[/yellow]] Enter a name for the new test suite, "
               "ensuring that it is non-empty and includes only alphanumeric "
               "characters, underscores, or dashes. Your test suites must "
               "have unique names.")
    return inquirer.text(
        message="Test suite name",
        validate=validate_test_suite_name)


def get_test_suite_description() -> Optional[str]:
    """Prompt user for a test suite description.
    
    Returns:
        Test suite description, or None if the user skipped this question.
    """
    rich.print("\n[[yellow]3/4[/yellow]] Optionally enter a description for "
               "the new test suite. Press enter to skip.")
    description = inquirer.text(message="Test suite description")
    if not description:
        return None
    return description


def get_llm_program_fully_qualified_name() -> str:
    """Prompt user for a LLM program fully qualified name.

    Prompt the user to enter a LLM program fully qualified name. Validate the
    name (see `validate_llm_program` function below). If the name is valid,
    return it. If the name is invalid, continue prompting the user until a
    valid name is entered.

    Returns:
        LLM program fully qualified name.
    """
    def validate_llm_program(_: Dict[str, Any], current: str) -> bool:
        """Validate LLM program fully qualified name.

        A LLM program fully qualified name is valid if it:
        1. Is in the format <fully qualified module name>:<fully qualified
            object name>.
        2. Can instantiate a `data_model.LazyCallable` object.
        3. The instantiated object can import the underlying LLM program
            and get its parameter keys with `get_parameter_keys`.

        Args:
            _: Dictionary containing the answers from previous questions
                (not including the current question). This will be an empty
                dictionary because the `llm_program_fully_qualified_name`
                question is not part of a collection of questions.
            current: The current answer, which is a LLM program fully
                qualified name to validate.

        Returns:
            True if the LLM program fully qualified name is valid, False
            otherwise.

        Raises:
            inquirer.errors.ValidationError: If the LLM program fully
                qualified name is invalid.
        """
        try:
            if not re.match(r".+:.+", current):
                raise inquirer.errors.ValidationError(
                    "",
                    reason="Invalid format.")

            # A valid LLM program fully qualified name should be able to
            # instantiate a `data_model.LazyCallable` object, which will
            # attempt to access the underlying callable object of the LLM
            # program. If it is unable to do so, it will raise an exception.
            data_model.LazyCallable(current)
        except inquirer.errors.ValidationError as error:
            raise error
        except Exception as error:
            # It is not trival identify the relevant Exception subtypes that
            # should be considered actual validation errors, so we catch all
            # exceptions and raise a generic validation error with the
            # exception message appended. The issue with this method is that
            # the validation error message is limited to a single line. If
            # the exception message is too long, it will be truncated. In
            # testing, exception messages seem to be able to fit on a single
            # line, but this should be revisited if we are encountering
            # exceptions with long messages.
            raise inquirer.errors.ValidationError(  # pylint: disable=raise-missing-from
                "", reason=f"Invalid object: {str(error)}")
        return True

    rich.print("\n[[yellow]4/4[/yellow]] Enter the fully qualified name (FQN) "
               "of the LLM program to test. The LLM program can be a Python "
               "function or a LangChain chain. The FQN should be in the "
               "format `path.to.python.module:function_or_object_name`.")
    return inquirer.text(
        message=("LLM program FQN"),
        validate=validate_llm_program)


def choose_new_test_suite() -> bool:
    """Prompt user to create a new test suite or use an existing one.

    Returns:
        True if the user wants to create a new test suite, False if the user
        wants to use an existing test suite.
    """
    create_new, use_existing = ("Create new", "Use existing")
    answer = inquirer.list_input(
        "Create new or use existing test suite?",
        choices=[create_new, use_existing])
    if answer == create_new:
        return True
    return False



# Copyright 2023 Inductor, Inc.
"""Functions for interfacing with the Inductor backend."""
# TODO: Add --verbose mode for printing out the request and response.

import json
from typing import Any, Dict

import httpx

from inductor import config, wire_model


# TODO: Cleanup these globals.
_MOCK_ENDPOINTS = False
_MOCK_USER_ID = {"user_id": 1}
_settings = config.Settings()
# _settings.inductor_api_url = "http://127.0.0.1:8000"


def _post_request(
    endpoint: str, request_body: Dict[str, Any], auth_access_token: str
) -> httpx.Response:
    """POST request to the backend.

    Args:
        endpoint: The endpoint to POST to. Should not begin with a slash.
        request: The request to POST.
        auth_access_token: Auth0 access token.

    Returns:
        The response from the backend.

    Raises:
        RuntimeError: If the response status code is not a successful response
            (i.e., if the response status code is not in the range (200 – 299)).
    """
    response = httpx.post(
        f"{_settings.inductor_api_url}/api/client/v1/{endpoint}",
        headers={"Authorization": f"Bearer {auth_access_token}"},
        params=_MOCK_USER_ID,
        json=json.loads(request_body.model_dump_json()),
    )
    # TODO: These print statements should be avaialbe in --verbose mode.
    # print(f"Response: {response.text}")
    # print(f"Status code: {response.status_code}")
    # print(f"JSON: {response.json()}")
    if response.status_code < 200 or response.status_code >= 300:
        raise RuntimeError(
            f"POST request to `{endpoint}` failed with status code "
            f"{response.status_code} and response text: {response.text}."
        )
    return response


def create_api_key(
    request: wire_model.CreateApiKeyRequest,
    auth_access_token: str):
    """POST request to create an API key.
    
    Args:
        request: CreateApiKeyRequest object
        auth_access_token: Auth0 access token.
    """
    if _MOCK_ENDPOINTS:
        return
    _post_request("create-api-key", request, auth_access_token)


def create_test_suite(
    request: wire_model.CreateTestSuiteRequest,
    auth_access_token: str
) -> wire_model.CreateTestSuiteResponse:
    """POST request to create a test suite.

    Args:
        request: CreateTestSuiteRequest object
        auth_access_token: Auth0 access token.
    
    Returns:
        CreateTestSuiteResponse object
    """
    if _MOCK_ENDPOINTS:
        return wire_model.CreateTestSuiteResponse(id=123)
    response = _post_request("create-test-suite", request, auth_access_token)
    return wire_model.CreateTestSuiteResponse(**response.json())


def create_test_suite_run(
    request: wire_model.CreateTestSuiteRunRequest,
    auth_access_token: str
) -> wire_model.CreateTestSuiteRunResponse:
    """POST request to create a test suite run.
    
    Args:
        request: CreateTestSuiteRunRequest object
        auth_access_token: Auth0 access token.
    
    Returns:
        CreateTestSuiteRunResponse object
    """
    if _MOCK_ENDPOINTS:
        return wire_model.CreateTestSuiteRunResponse(
            test_suite_run_id=123,
            test_case_ids=[1, 2, 3, 4, 5, 6],
            quality_measure_ids=[7, 8, 9, 10, 11, 12],
            hparam_spec_ids=[13, 14, 15, 16, 17, 18],
            url="http://localhost:5000/MOCK_URL",
        )
    response = _post_request(
        "create-test-suite-run", request, auth_access_token)
    return wire_model.CreateTestSuiteRunResponse(**response.json())


def log_test_case_execution(
    request: wire_model.LogTestCaseExecutionRequest,
    auth_access_token: str):
    """POST request to log test case execution.

    Args:
        request: LogTestCaseExecutionRequest object
        auth_access_token: Auth0 access token.
    """
    if _MOCK_ENDPOINTS:
        return
    _post_request("log-test-case-execution", request, auth_access_token)


def complete_test_suite_run(
    request: wire_model.CompleteTestSuiteRunRequest,
    auth_access_token: str):
    """POST request to complete test suite run.

    Args:
        request: CompleteTestSuiteRunRequest object
        auth_access_token: Auth0 access token.
    """
    if _MOCK_ENDPOINTS:
        return
    _post_request("complete-test-suite-run", request, auth_access_token)


def log_llm_program_execution_request(
    request: wire_model.LogLlmProgramExecutionRequest,
    auth_access_token: str):
    """POST request to log LLM program execution.

    Args:
        request: LogLlmProgramExecutionRequest object
        auth_access_token: Auth0 access token.
    """
    if _MOCK_ENDPOINTS:
        return
    _post_request("log-llm-program-execution", request, auth_access_token)


# def get_existing_test_suites() -> List[str]:
#     """TODO"""
#     existing_test_suites = ["test_suite_1", "test_suite_2"]
#     return existing_test_suites



# Copyright 2023 Inductor, Inc.
"""Entrypoint for the Inductor CLI."""

import contextlib
import datetime
import textwrap
from typing import Annotated, Any, Dict, List, Optional
from pathlib import Path

import inquirer
import loky
import rich
from rich import console, progress, syntax
import typer
import yaml

from inductor.cli import data_model, execute, questions
from inductor import auth_session, backend_client, config, wire_model


app = typer.Typer()


def _create_test_suite_file(
    auth_access_token: str,
    default_vars: Optional[Dict[str, Any]] = None,):
    """Prompt user for test suite info to create a test suite file.

    1. Prompt user to create a new test suite or use an existing one.
    2. If creating a new test suite, prompt user for test suite info.
    3. If using an existing test suite, prompt user to select one.
    4. Create a test suite file with the provided info.
    
    Args:
        auth_access_token: Auth0 access token.
        default_vars: Optional dictionary containing default values for
            populating the test suite file. Only items with keys that are
            in the `model.TestSuiteRun` dataclass will be used.
    """
    if default_vars is None:
        default_vars = {}
    test_suite_args = data_model.process_test_suite_args(default_vars)

    create_new_test_suite = questions.choose_new_test_suite()

    if create_new_test_suite:
        rich.print("Creating a new test suite file ...")
        test_suite_file_path = questions.get_new_test_suite_file_path()
        test_suite_args["name"] = questions.get_test_suite_name()
        test_suite_args["description"] = questions.get_test_suite_description()

        if test_suite_args.get("llm_program_fully_qualified_name") is None:
            test_suite_args["llm_program_fully_qualified_name"] = (
                questions.get_llm_program_fully_qualified_name())

        test_suite_args["id"] = None

    else:
        # TODO: This `else` block is a placeholder until the backend is
        # implemented.
        # existing_test_suites = backend_client.get_existing_test_suites()
        # inquirer.list_input(
        #     "Select test suite to use", choices=existing_test_suites)
        # existing_test_suite_args = {}
        # test_suite_args["llm_program_fully_qualified_name"] = "module:object"
        # test_suite_args["id"] = "123"
        # test_suite_args = data_model.merge_test_suite_args(
        #     priority_1_dict=test_suite_args,
        #     priority_2_dict=existing_test_suite_args)
        # raise NotImplementedError("Use existing test suite not implemented.")
        rich.print("To use an existing test suite, run `inductor test "
                   "\[TEST_SUITE_FILE_PATH]`.")  # pylint: disable=anomalous-backslash-in-string
        raise typer.Exit()

    # Get parameter keys from LLM program to populate
    # the test suite file's test case section.
    llm_program = data_model.LazyCallable(
        test_suite_args["llm_program_fully_qualified_name"])
    llm_program_param_keys = llm_program.inputs_signature.keys()
    inputs = {}
    for key in llm_program_param_keys:
        inputs[key] = "<REPLACE WITH VALUE>"

    # Create test suite file contents.
    newline = "\n"
    yaml_string = textwrap.dedent(f"""
    # Test suite file

    # Test suite configuration
    - config:
        id: <WILL BE POPULATED AUTOMATICALLY>
        # name: {test_suite_args["name"]}
        # description: {
            test_suite_args["description"]
            if test_suite_args["description"] is not None
            else ""}
        llm_program_fully_qualified_name: {
            test_suite_args["llm_program_fully_qualified_name"]}
        replicas: {test_suite_args.get("replicas", 1)}
        
    # A test case (add more `- test:` blocks to add more test cases)
    - test:
        inputs:
          {yaml.dump(inputs, sort_keys=False).strip().replace(
            newline, newline + "          ")}
    
    # Sample quality measure (add more `- quality:` blocks to add more
    # quality measures)
    # - quality:
    #     name: <REPLACE WITH HUMAN-READABLE NAME FOR QUALITY MEASURE>
    #     evaluator: <REPLACE WITH "FUNCTION", "HUMAN", OR "LLM">
    #     evaluation_type: <REPLACE WITH "BINARY" OR "RATING_INT">
    #     # If evaluator is "FUNCTION", spec should be the fully qualified
    #     # name of your quality measure function (e.g., "my.module:my_function").
    #     # If evaluator is "HUMAN", spec should be the instruction or question
    #     # to be posed to a human evaluator. If evaluator is "LLM", spec should
    #     # be in the form of model and prompt (e.g. model: "gpt-3.5-turbo" and
    #     # prompt: "...") or the fully qualified name of an LLM program to be
    #     # used as the evaluator (e.g., "my.module:my_function").  Please see
    #     # the Inductor docs for more details.
    #     spec: <REPLACE WITH VALUE>

    # Sample hyperparameter specification (add more `- hparam:` blocks to add more
    # hyperparameters)
    # - hparam:
    #     name: <REPLACE WITH STRING-VALUED NAME FOR HYPERPARAMETER>
    #     type: <REPLACE WITH "SHORT_STRING", "TEXT", OR "NUMBER">
    #     values: <REPLACE WITH LIST OF VALUES FOR HYPERPARAMETER>
    """).lstrip()

    rich.print("\nTest suite file path:")
    rich.print(test_suite_file_path)
    rich.print("\nTest suite file contents:")
    yaml_string_with_syntax = syntax.Syntax(
        yaml_string, "yaml", theme="ansi_dark", line_numbers=True)
    rich_console = console.Console()
    rich_console.print(yaml_string_with_syntax)

    # Write the test suite file.
    confirm = inquirer.confirm(
        "Create test suite and write contents?",
        default=True)
    if confirm:
        # Create test suite on server if it does not already exist.
        if test_suite_args.get("id") is None:
            response = backend_client.create_test_suite(
                wire_model.CreateTestSuiteRequest(
                    name=test_suite_args["name"],
                    description=test_suite_args.get("description")),
                auth_access_token
            )
            test_suite_args["id"] = response.id
        rich.print("Test suite created.")
        # Replace the `id` field in the test suite file with the ID of the
        # test suite created on the server.
        yaml_string = yaml_string.replace(
            "<WILL BE POPULATED AUTOMATICALLY>", str(test_suite_args["id"]))

        # Write the test suite file.
        with open(test_suite_file_path, "w") as file:
            file.write(yaml_string)
        rich.print("Next, add test cases and quality measures to the file, "
                   "and run your test suite via `inductor test "
                   f"{test_suite_file_path}`")


@contextlib.contextmanager
def _manage_test_suite_run(
    test_suite_run: wire_model.CreateTestSuiteRunRequest,
    auth_access_token: str) -> wire_model.CreateTestSuiteRunResponse:
    """Send requests to the server to manage the creation/completion of a run.
    
    Send a request to the server to create a test suite run. Then, yield the
    response, which contains test suite run metadata. On exit, send a request
    to the server to mark the test suite run as complete.

    Args:
        test_suite_run: CreateTestSuiteRunRequest object.
        auth_access_token: Auth0 access token.

    Yields:
        CreateTestSuiteRunResponse object.
    """
    test_suite_run_metadata = backend_client.create_test_suite_run(
        test_suite_run, auth_access_token)
    try:
        yield test_suite_run_metadata
    finally:
        backend_client.complete_test_suite_run(
            wire_model.CompleteTestSuiteRunRequest(
                test_suite_run_id=
                    test_suite_run_metadata.test_suite_run_id,
                ended_at=
                    datetime.datetime.now(datetime.timezone.utc)),
            auth_access_token)


@app.command("test")
def test(
    test_suite_file_paths: Annotated[Optional[List[Path]], typer.Argument(help=(
        "One or more paths to test suites to run, separated by spaces."
    ))] = None,
    llm_program_fully_qualified_name: Annotated[Optional[str], typer.Option(  # pylint: disable=unused-argument
        "-l",
        "--llm_program",
        help=(
            "Fully qualified name of the Python object (in particular, a "
            "Python function or LangChain chain) representing this LLM "
            "program, in the format: "
            "`<fully qualified module name>:<fully qualified object name>` "
            "(e.g., `my.module:object`). <fully qualified module name> can be "
            "in the format: `path.to.module` or `path/to/module.py`."
    ))] = None,
    replicas: Annotated[Optional[int], typer.Option(  # pylint: disable=unused-argument
        "-r",
        "--replicas",
        help="Number of runs to execute for each test case in test suite(s)."
    )] = None,
    parallelize: Annotated[Optional[int], typer.Option(  # pylint: disable=unused-argument
        "-p",
        "--parallelize",
        help="Maximum number of test cases to run simultaneously in parallel."
    )] = None,
    # TODO: Add support for imports.
    # imports: Annotated[Optional[List[Path]], typer.Option(  # pylint: disable=unused-argument
    #     "-i",
    #     "--imports",
    #     help=("One or more paths to files containing additional test cases "
    #           "or quality measures to include in the test suite(s).")
    # )] = None,
    verbose: Annotated[Optional[bool], typer.Option(
        "-v",
        "--verbose",
        help="Print verbose output.",
    )] = False,
):
    """Run one or more Inductor test suites."""
    local_vars = locals()

    # TODO: Add verbose flag for the main CLI application.
    config.verbose = verbose

    # Every time the access token is requested from the session, it is
    # refreshed if it is expired.
    auth_access_token = auth_session.get_auth_session().access_token

    if not test_suite_file_paths:
        _create_test_suite_file(auth_access_token, local_vars)
        raise typer.Exit()
    test_suite_runs = data_model.get_test_suite_runs(local_vars)
    multiple_runs = len(test_suite_runs) > 1

    for i, test_suite_run in enumerate(test_suite_runs):

        if multiple_runs:
            rich.print(rich.rule.Rule(
                f"Test Suite {i+1} of {len(test_suite_file_paths)}",
                style="bold #ffffff"))

        def pluralize(count: int) -> str:
            return "s" if count != 1 else ""
        rich.print(f"Running {len(test_suite_run.test_cases)} test case"
              f"{pluralize(len(test_suite_run.test_cases))} "
              f"({test_suite_run.replicas} run"
              f"{pluralize(test_suite_run.replicas)} each) ...")

        execute_test_suite_run(
            test_suite_run,
            auth_access_token,
            prompt_open_results=not multiple_runs)


def execute_test_suite_run(
    test_suite_run: wire_model.CreateTestSuiteRunRequest,
    auth_access_token: str,
    *,
    prompt_open_results: bool = False):
    """Execute a test suite run.
    
    Execute a test suite run while displaying relevant information to the user,
    including a progress bar.
    
    Args:
        test_suite_run: CreateTestSuiteRunRequest object.
        auth_access_token: Auth0 access token.
        prompt_open_results: Whether to prompt the user to open the test
            suite run results in a browser.
    """
    with (
        loky.get_reusable_executor(
            max_workers=test_suite_run.parallelize) as executor,
        _manage_test_suite_run(
            test_suite_run, auth_access_token) as test_suite_run_metadata
    ):
        test_case_futures = execute.execute_test_suite_run(
            test_suite_run,
            test_suite_run_metadata,
            executor,
            auth_access_token)

        rich.print(f"Go to {test_suite_run_metadata.url} to view results.")
        if prompt_open_results:
            open_url = inquirer.confirm(
                message="Open in browser?", default=True)
            if open_url:
                typer.launch(test_suite_run_metadata.url)

        # Display progress bar and optionally print test outputs.
        with progress.Progress() as progress_bar:
            progress_task = progress_bar.add_task(
                "Test Cases",
                total=(test_suite_run.replicas *
                        len(test_suite_run.test_cases) *
                        len(execute.get_hparams_combinations(
                            test_suite_run.hparam_specs))))

            for future in loky.as_completed(test_case_futures):
                progress_bar.advance(progress_task)
                test_output, invalid_quality_measures = future.result()
                if test_output.execution_details.error is not None:
                    progress_bar.console.print(
                        "\n[red][bold][ERROR] Test case execution "
                        "raised an exception.[/bold] The following "
                        "execution will be recorded as FAILED and "
                        "quality measures will not be evaluated:[/red]")
                    progress_bar.console.print(test_output)
                elif config.verbose:
                    progress_bar.console.print(test_output)
                if invalid_quality_measures:
                    progress_bar.console.print(
                        "\n[red][bold][ERROR] One or more quality "
                        "measures raised an exception or returned an "
                        "invalid value.[/bold] The following quality "
                        "measures will not be recorded as part of this "
                        "test case execution:[/red]")
                    progress_bar.console.print(invalid_quality_measures)

        rich.print("Run complete.")


# Required to enable subcommands with only one command.
# For more information, see:
# https://typer.tiangolo.com/tutorial/commands/one-or-multiple/#one-command-and-one-callback
# NOTE: This function should be removed when we add a second command.
# NOTE: The above is written as a comment, rather than a docstring, to prevent
#       it from displaying when the CLI is run with the --help flag.
@app.callback()
def callback():  # pylint: disable=missing-function-docstring
    pass



# Copyright 2023 Inductor, Inc.
"""Types used by backend API that powers Inductor's client (CLI and library)."""

import datetime
from typing import Any, Callable, Dict, List, Literal, Optional, Union

import pydantic

from inductor import util


# Field definitons
_RESTRICTED_NAME_FIELD = pydantic.Field(
    max_length=1000, pattern=r"^[a-zA-Z0-9_-]+$")
_SUBRESOURCE_NAME_FIELD = pydantic.Field(max_length=500)


class CreateApiKeyRequest(pydantic.BaseModel):
    """Request body type for create API key endpoint.

    Attributes:
        auth0_id: Auth0 ID of the API key.
    """
    auth0_id: str


class TestCase(pydantic.BaseModel):
    """A test case.

    Attributes:
        inputs: Mapping from input parameter name to input value.
        output: Optionally, an example of a desired high-quality output, or
            the output that is to be considered correct, for this test case.
        description: Optionally, a description of this test case.
    """
    model_config = pydantic.ConfigDict(extra="forbid")

    inputs: Dict[str, Any]
    output: Optional[Any] = None
    description: Optional[str] = None

    # Define a custom constructor to enable taking `inputs` as a positional,
    # rather than keyword-only, argument.
    def __init__(
        self,
        inputs: Dict[str, Any],
        *,
        output: Optional[Any] = None,
        description: Optional[str] = None):
        """Constructs a new test case.

        Args:
            inputs: Mapping from input parameter name to input value.
            output: Optionally, an example of a desired high-quality output, or
                the output that is to be considered correct, for this test case.
            description: Optionally, a description of this test case.
        """
        super().__init__(inputs=inputs, output=output, description=description)


class QualityMeasure(pydantic.BaseModel):
    """A quality measure.
    
    Attributes:
        name: Human-readable name of this quality measure.
        evaluator: Evaluator for this quality measure.  Determines whether
            this quality measure will be evaluated by running a function,
            or via human inspection.
        evaluation_type: The type of value produced by evaluation of this
            quality measure.
        spec: Specification of the details of how to execute this quality
            measure.
            - If evaluator is "FUNCTION", then spec should give the fully
            qualified name of the function (in the format
            "my.module:my_function").
            - If evaluator is "HUMAN", then spec should give the instructions
            or question that should be displayed to human evaluators.
            - If evaluator is "LLM", then spec should either give a dictionary
            with "model" and "prompt" fields, or the fully qualified name of an
            LLM program that implements the quality measure (in the format
            "my.module:my_function").
    """
    model_config = pydantic.ConfigDict(extra="forbid")

    name: str = _SUBRESOURCE_NAME_FIELD
    evaluator: Literal["FUNCTION", "HUMAN", "LLM"]
    evaluation_type: Literal["BINARY", "RATING_INT"]
    spec: Union[str, Dict[str, str]]

    @pydantic.field_validator("spec", mode="before")
    @classmethod
    def extract_fully_qualified_name(cls, v: Union[str, Callable]):
        """Return the fully qualified name from the spec, if applicable."""
        if callable(v):
            return f"{util.get_module_qualname(v)}:{v.__qualname__}"
        return v


class HparamSpec(pydantic.BaseModel):
    """Specification of set of hyperparameter values to use for test suite run.

    Attributes:
        hparam_name: Name of hyperparameter.
        hparam_type: Type of hyperparameter.
        values: List of hyperparameter values.
    """
    model_config = pydantic.ConfigDict(extra="forbid", populate_by_name=True)

    hparam_name: str = pydantic.Field(max_length=500, alias="name")
    hparam_type: Literal["SHORT_STRING", "TEXT", "NUMBER"] = pydantic.Field(
        alias="type")
    values: List[Any]


class ProgramDetails(pydantic.BaseModel):
    """Details of an LLM program.

    Attributes:
        fully_qualified_name: Fully qualified name of the LLM program.
        inputs_signature:  Map between input parameter names to strings
            indicating the corresponding parameter types (or to null for
            input parameters that do not have type annotations).
        program_type: Type of LLM program.
    """
    fully_qualified_name: str
    inputs_signature: Dict[str, Optional[str]]
    program_type: Literal["FUNCTION", "LANGCHAIN"]


class LoggedValue(pydantic.BaseModel):
    """A logged value associated with an LLM program execution.

    Attributes:
        value: The logged value.
        description: Description of the logged value (if any).
        after_complete: Whether the logged value was logged after the LLM
            program completed (as opposed to during its execution).
    """
    value: Any
    description: Optional[str] = None
    after_complete: bool


class DirectEvaluation(pydantic.BaseModel):
    """A direct evaluation of a quality measure.
    
    Attributes:
        quality_measure_id: ID of the quality measure that was evaluated.
        value_bool: The output of the quality measure, if boolean.
        value_int: The output of the quality measure, if an integer.
    """
    quality_measure_id: int
    value_bool: Optional[bool] = None
    value_int: Optional[int] = None


# TODO: This class is currently not used to transmit data to the backend, but
# instead is a placeholder until the backend supports recording quality measure
# execution details (i.e. errors, stdout, and stderr).
class QualityMeasureExecutionDetails(pydantic.BaseModel):
    """Details of a quality measure execution.

    Attributes:
        input: The input to the quality measure.
        output: The output of the quality measure, if available.  Should be
            None if the execution terminated with an error.
        error: If an error occurred during quality measure execution, this
            field should contain a specification of the error.
        stdout: The content printed to stdout during the quality measure's
            execution.
        stderr: The content printed to stderr during the quality measure's
            execution.
    """
    input: Any
    output: Optional[Any] = None
    error: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None


class ExecutionDetails(pydantic.BaseModel):
    """An LLM program execution.

    Attributes:
        mode: Mode in which this execution was performed (e.g., via CLI, or as
            part of a deployment), if known.
        inputs: JSON-serializable mapping from input argument name to input
            value.
        hparams: JSON-serializable mapping from hyperparameter name to
            hyperparameter value.
        output: The output of the LLM program, if available.  Should be None
            if the execution terminated with an error, without producing an
            output.
        error: If an error occurred during LLM program execution, this column
            should contain a specification of the error.
        stdout: The content printed to stdout during the LLM program's
            execution.
        stderr: The content printed to stderr during the LLM program's
            execution.
        execution_time_secs: The total wall clock time elapsed during this
            execution.
        started_at: The timestamp at which the execution started.
        ended_at: The timestamp at which the execution ended (whether
            successfully or due to an error).
        logged_values: The values logged during the LLM program's execution.
        direct_evaluations: The direct evaluations performed during the LLM
            program's execution.
    """
    mode: Literal["CLI", "DEPLOYED"]
    inputs: Dict[str, Any]
    hparams: Optional[Dict[str, Any]] = None
    output: Optional[Any] = None
    error: Optional[str] = None
    stdout: Optional[str] = None
    stderr: Optional[str] = None
    execution_time_secs: float
    started_at: datetime.datetime
    ended_at: datetime.datetime
    logged_values: Optional[List[LoggedValue]] = None
    direct_evaluations: Optional[List[DirectEvaluation]] = None


class CreateTestSuiteRequest(pydantic.BaseModel):
    """Request body for create test suite endpoint.

    Attributes:
        name: Name of test suite. Test suite names must be unique per user.
        description: Description of test suite.
    """
    name: str = _RESTRICTED_NAME_FIELD
    description: Optional[str] = None


class CreateTestSuiteResponse(pydantic.BaseModel):
    """Response for create test suite endpoint upon successful creation.
    
    Attributes:
        id: ID of the created test suite.
    """
    id: int


class CreateTestSuiteRunRequest(pydantic.BaseModel):
    """Request body for create test suite run endpoint.

    Attributes:
        test_suite_id_or_name: ID or name of test suite to run.
        test_cases: List of test cases.
        quality_measures: List of quality measures (if any).
        hparam_specs: List of hyperparameter specifications (if any).
        llm_program_details: Details of LLM program to run.
        replicas: Number of times that LLM program will be run on each pair of
            (test case, set of hyperparameters).
        parallelize: Degree of parallelism used for this run.
        started_at: The timestamp at which the test suite run started.
    """
    test_suite_id_or_name: Union[int, str]
    test_cases: List[TestCase]
    quality_measures: Optional[List[QualityMeasure]] = None
    hparam_specs: Optional[List[HparamSpec]] = None
    llm_program_details: ProgramDetails
    replicas: int = 1
    parallelize: int = 1
    started_at: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc))


class CreateTestSuiteRunResponse(pydantic.BaseModel):
    """Response for create test suite run endpoint upon successful creation.
    
    Attributes:
        test_suite_run_id: ID of the created test suite run.
        test_case_ids: IDs of the created test cases. The order of test case
            IDs is guaranteed to be the same as the order of test cases in the
            request.
        quality_measure_ids: IDs of the created quality measures. The order of
            quality measure IDs is guaranteed to be the same as the order of
            quality measures in the request.
        hparam_spec_ids: IDs of the created hyperparameter specifications. The
            order of hyperparameter specification IDs is guaranteed to be the
            same as the order of hyperparameter specifications in the request.
        url: URL at which the created test suite run can be accessed.
    """
    test_suite_run_id: int
    test_case_ids: List[int]
    quality_measure_ids: List[int]
    hparam_spec_ids: List[int]
    url: str


class LogTestCaseExecutionRequest(pydantic.BaseModel):
    """Request body for log test case execution endpoint.

    Attributes:
        test_suite_run_id: ID of test suite run as part of which this
            execution occurred (if any).
        test_case_id: ID of test case on which this execution occurred.
        test_case_replica_index: Index of execution replica on test case given
            by test_case_id. Replica indices should start at zero.
        execution_details: Details of the test case execution.
    """
    test_suite_run_id: int
    test_case_id: int
    test_case_replica_index: int
    execution_details: ExecutionDetails


class CompleteTestSuiteRunRequest(pydantic.BaseModel):
    """Request body for complete test suite run endpoint.

    Contains all remaining fields required by TestSuiteRun backend data model
    that have not already been logged.

    Attributes:
        test_suite_run_id: ID of the test suite run to complete.
        ended_at: The timestamp at which the test suite run ended.
    """
    test_suite_run_id: int
    ended_at: datetime.datetime


class LogLlmProgramExecutionRequest(pydantic.BaseModel):
    """Request body for log LLM program execution endpoint.

    Attributes:
        program_details: Details of the LLM program.
        execution_details: Details of the LLM program execution.
    """
    program_details: ProgramDetails
    execution_details: ExecutionDetails



# Copyright 2023 Inductor, Inc.
"""Inductor client configuration."""

from typing import Optional

import pydantic_settings


# TODO: Add levels of verbosity and possibly integrate it with a logger.
verbose = False


class Settings(pydantic_settings.BaseSettings):
    """Settings for the Inductor client.
    
    Settings are loaded from the following sources, in order of increasing
    precedence:
    1. Initialized values in this class
    2. Environment variables
    
    Attributes:
        auth0_domain: The Auth0 domain.
        auth0_client_id: The Auth0 client ID.
        inductor_api_url: The Inductor API URL.
        inductor_secret_key: The user's Inductor secret key.  If this is set,
            the client will use this secret key for authentication instead of
            prompting the user to authenticate.
        secret_key_prefix: The prefix for the secret key.
    """

    auth0_domain: str = "login.inductor.ai"
    auth0_client_id: str = "YFeXBAUixEVB3e9sM5e7XDZr4FEHH5XX"
    inductor_api_url: str = "https://app.inductor.ai"
    inductor_secret_key: Optional[str] = None
    secret_key_prefix: str = "isk_"

    model_config = pydantic_settings.SettingsConfigDict(extra="ignore")



# Copyright 2023 Inductor, Inc.
"""Abstractions for the Inductor CLI."""

import copy
import importlib
import inspect
import os
from pathlib import Path
import sys
import types
from typing import Any, Callable, Dict, List, Optional, Union

import yaml

from inductor import util, wire_model


class LazyCallable:
    """Container for a function or LangChain object that is lazily imported.
    
    Represents a function or LangChain object that is provided by the user as
    part of a test suite. Objects of this class are callable, and calling them
    will import the required module if necessary to call the object.

    Attributes:
        fully_qualified_name (str): Fully qualified name of the callable
            object, in the format "<fully qualified module name>:<fully
            qualified object name>". (e.g. "my.module:my_function")
        path_to_module_dir (str): Path to the directory that contains the
            module that contains the callable object.
        input_signature (Dict[str, Optional[str]]): Inputs signature of the
            callable object. Inputs signature is a map between input parameter
            names to strings indicating the corresponding parameter types (or to
            None for input parameters that do not have type annotations).
    """
    def __init__(
        self,
        callable_or_fully_qualified_name: Union[Callable, str],
        path_to_module_dir: str = os.getcwd()):
        """Initialize a LazyCallable object.

        Args:
            callable_or_fully_qualified_name: Callable object or fully
                qualified name of the callable object, in the format
                "<fully qualified module name>:<fully qualified object name>".
                (e.g. "my.module:my_function")
            path_to_module_dir: Path to the directory that contains the module
                that contains the callable object. If not provided, the
                current working directory is used. This argument is ignored if
                `callable_or_fully_qualified_name` is a callable object.
        """
        if isinstance(callable_or_fully_qualified_name, Callable):
            callable_object = callable_or_fully_qualified_name
            module_qualname = util.get_module_qualname(callable_object)
            object_qualname = callable_object.__qualname__

            self.fully_qualified_name = f"{module_qualname}:{object_qualname}"
            self.path_to_module_dir = os.path.dirname(
                inspect.getfile(callable_object))

        elif isinstance(callable_or_fully_qualified_name, str):
            self.fully_qualified_name = callable_or_fully_qualified_name
            self.path_to_module_dir = path_to_module_dir

            callable_object = self.get_callable()

        else:
            raise ValueError(
                "callable_or_fully_qualified_name must be a callable or a "
                "string.")

        if callable_object.__class__.__name__ == "LLMChain":
            self._program_type = "LANGCHAIN"
        elif inspect.isfunction(callable_object):
            self._program_type = "FUNCTION"
        else:
            raise ValueError(
                f"Object {self.fully_qualified_name} is not a function or "
                "LangChain object.")

        self.inputs_signature = self._get_inputs_signature()

    def get_details(self) -> wire_model.ProgramDetails:
        """Return a `wire_model.ProgramDetails` object.
        
        Returns:
            `wire_model.ProgramDetails` object.
        """
        return wire_model.ProgramDetails(
            fully_qualified_name=self.fully_qualified_name,
            program_type=self._program_type,
            inputs_signature=self.inputs_signature)

    def _import_module(self, module_qualname: str) -> types.ModuleType:
        """Import the module given by module_qualname.

        Args:
            module_qualname: Fully qualified name of the module to import.

        Returns:
            The imported module.
        """
        orig_sys_path = sys.path.copy()
        sys.path[0] = self.path_to_module_dir
        module = importlib.import_module(module_qualname)
        sys.path = orig_sys_path
        return module

    def get_callable(self) -> Callable:
        """Import the callable object and return it.
        
        Returns:
            Callable object.
        """
        module_qualname, object_qualname = (
            self.fully_qualified_name.split(":"))
        module_qualname = module_qualname.replace(
            "/", ".").removesuffix(".py")
        if module_qualname == "google_colab":
            module_qualname = "__main__"

        module = self._import_module(module_qualname)
        callable_object = module
        for name in object_qualname.split("."):
            callable_object = getattr(callable_object, name)
        return callable_object

    def __call__(self, *args, **kwargs):
        """Call the callable object.
        
        Args:
            *args: Positional arguments to pass to the callable object.
            **kwargs: Keyword arguments to pass to the callable object.
        """
        callable_object = self.get_callable()
        if self._program_type == "LANGCHAIN":
            return callable_object.run(*args, **kwargs)
        else:
            return callable_object(*args, **kwargs)

    def _get_inputs_signature(self) -> Dict[str, Optional[str]]:
        """Return the inputs signature of the callable object.
        
        Inputs signature is a map between input parameter names to
        strings indicating the corresponding parameter types (or to None for
        input parameters that do not have type annotations).
        """
        callable_object = self.get_callable()
        if self._program_type == "LANGCHAIN":
            return {key: None for key in callable_object.input_keys}
        else:
            signature = inspect.signature(callable_object)
            return {
                name: (
                    str(param.annotation)
                    if param.annotation != inspect._empty  # pylint: disable=protected-access
                    else None)
                for name, param in signature.parameters.items()
            }


def get_test_suite_runs(
    local_vars: Dict[str, Any]) -> List[wire_model.CreateTestSuiteRunRequest]:
    """Return a list of test suite runs (which have not yet been executed).

    From the given dictionary, extract the relevant command line arguments,
    including the test suite file paths (which are required and contain
    additional configuration arguments). Create TestSuiteRun instances based on
    the specified hierarchy of configuration sources:
    1. Command line arguments.
    2. Test suite file arguments.
    3. Default arguments.
    In the case of a conflict, the lower-numbered item in the list takes
    precedence.
    
    Args:
        local_vars: Dictionary of local variables used to populate
            `wire_model.CreateTestSuiteRunRequest`. Only items with
            keys that are in `wire_model.CreateTestSuiteRunRequest`
            will be read.

    Returns:
        A list of `wire_model.CreateTestSuiteRunRequest` objects.

    Raises:
        ValueError: If the `test_suite_file_paths` argument is not defined in
            the given dictionary.
    """
    test_suite_file_paths = local_vars.get("test_suite_file_paths")
    if test_suite_file_paths is None:
        raise ValueError("test_suite_file_paths must be defined.")

    test_suite_runs = []
    for test_suite_file_path in test_suite_file_paths:
        # Process command line arguments and test suite file arguments
        # separately. Then merge them, with command line arguments taking
        # precedence.
        test_suite_cli_args = process_test_suite_args(local_vars)
        test_suite_file_args = _process_test_suite_file(
            test_suite_file_path, includes_config=True)
        test_suite_args = merge_test_suite_args(
            priority_1_dict=test_suite_cli_args,
            priority_2_dict=test_suite_file_args)

        # Process imports and merge them with the other arguments.
        for file in test_suite_args.get("imports", []):
            additional_args = _process_test_suite_file(
                file, includes_config=False)
            test_suite_args = merge_test_suite_args(
                priority_1_dict=test_suite_args,
                priority_2_dict=additional_args)

        # Create ProgramDetails object.
        test_suite_args["llm_program_details"] = LazyCallable(
            test_suite_args.pop("llm_program_fully_qualified_name")
        ).get_details()

        test_suite_runs.append(
            wire_model.CreateTestSuiteRunRequest(**test_suite_args))

    return test_suite_runs


def merge_test_suite_args(
    priority_1_dict: Dict[str, Any],
    priority_2_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Create a new dictionary by combining the contents of two dictionaries.

    The priority 1 dictionary takes precedence over the priority 2 dictionary.
    Values that are lists in both dictionaries are combined, not overwritten.
    
    Args:
        priority_1_dict: Dictionary that takes precedence over the priority 2
            dictionary.
        priority_2_dict: Dictionary that is overwritten by the priority 1
            dictionary.

    Returns:
        New dictionary that combines the contents of the two dictionaries.
    """
    merged_dict = copy.deepcopy(priority_2_dict)
    for key, value in priority_1_dict.items():
        # Combine values if both are lists.
        if (key in merged_dict and
            isinstance(merged_dict[key], list) and
            isinstance(value, list)):
            merged_dict[key].extend(value)
        # Otherwise update the value.
        else:
            merged_dict[key] = value
    return merged_dict


def process_test_suite_args(
    args: Dict[str, Any],
    ignore_invalid_args: bool = True) -> Dict[str, Any]:
    """Return a dictionary of test suite run arguments.

    Given a dictionary of arguments, return a new dictionary that contains
    only the arguments that correspond to
    `wire_model.CreateTestSuiteRunRequest` and whose values are not None or
    empty lists.

    Args:
        args: Dictionary of arguments to process.
        ignore_invalid_args: Whether to ignore arguments that do not correspond
            to `wire_model.CreateTestSuiteRunRequest`. If False, an exception
            will be raised if an argument does not correspond to the
            `wire_model.CreateTestSuiteRunRequest`.

    Returns:
        Dictionary of test suite arguments.

    Raises:
        ValueError: If `ignore_invalid_args` is False and an argument does
            not correspond to `wire_model.CreateTestSuiteRunRequest`.
    """
    processed_args = {}
    for key, value in args.items():
        if key in wire_model.CreateTestSuiteRunRequest.__annotations__:
            if (value is not None and
                not (isinstance(value, list) and len(value) == 0)):
                processed_args[key] = value
        elif key == "id":
            processed_args["test_suite_id_or_name"] = value
        elif key == "imports":
            # TODO: Implement imports.
            raise NotImplementedError("Imports are coming soon!")
        elif key == "llm_program_fully_qualified_name":
            if value is not None:
                processed_args[key] = value
        elif key in ["name", "description"]:
            pass
        elif not ignore_invalid_args:
            raise ValueError(f"Unknown argument: {key}")
    return processed_args


def _process_test_suite_file(
    path: Path,
    *,
    includes_config: bool) -> Dict[str, Any]:
    """Return a dictionary of test suite arguments from a test suite file.
    
    Read a test suite file and return a dictionary of arguments that correspond
    to `wire_model.CreateTestSuiteRunRequest`.

    Args:
        path: Path to the test suite file.
        includes_config: Whether the test suite file is expected to include a
            config entry.

    Returns:
        Dictionary of test suite arguments.

    Raises:
        ValueError: If the test suite file contains more than one config entry
            or if the test suite file does not contain a config entry and
            `includes_config` is True.
    """
    with open(path, "r") as f:
        yaml_content = yaml.safe_load(f)

    test_suite_file_args = {
        "test_cases": [],
        "quality_measures": [],
    }
    config_found = False

    for entry in yaml_content:
        for key, value in entry.items():

            if key in ["test", "test case", "test_case"]:
                inputs=value.get("inputs")
                output=value.get("output")
                description=value.get("description")
                if inputs is None:
                    inputs = value
                    if any((output, description)):
                        raise ValueError(
                            "If including output or description in a test "
                            "case, inputs must be defined in an explicit "
                            "inputs block.")
                test_suite_file_args["test_cases"].append(
                    wire_model.TestCase(
                        inputs=inputs,
                        output=output,
                        description=description))

            elif key in [
                "quality", "quality measure", "quality_measure", "measure"]:
                quality_measure = wire_model.QualityMeasure(**value)
                test_suite_file_args["quality_measures"].append(
                    quality_measure)

            elif key in ["hparam", "hparam spec", "hparam_spec"]:
                # Disable the following field renaming for now, due to
                # modification of wire_model.HparamSpec's constructor to
                # support its direct usage within user-facing Python API.
                # for k in ["name", "type"]:
                #     value[f"hparam_{k}"] = value.pop(k)

                hparam_spec = wire_model.HparamSpec(**value)
                test_suite_file_args.setdefault("hparam_specs", []).append(
                    hparam_spec)

            elif key in ["config", "configuration"]:
                if includes_config and config_found:
                    raise ValueError(
                        "More than one config entry found in test suite file. "
                        "There must be exactly one config entry for each "
                        "test suite.")
                config_found = True
                if not includes_config:
                    raise ValueError(
                        "Only the primary test suite file may contain a "
                        "config entry.")
                config = process_test_suite_args(
                    value, ignore_invalid_args=False)
                test_suite_file_args.update(config)

            else:
                raise ValueError(f"Invalid entry: {key}")

    if not config_found and includes_config:
        raise ValueError(
            "No config entry found in test suite file. "
            "There must be exactly one config entry for each test suite.")

    return test_suite_file_args



# Copyright 2023 Inductor, Inc.
"""Inductor utility functions."""

import os
from typing import Callable


def get_module_qualname(f: Callable) -> str:
    """Return the fully qualified name of the module in which f is defined.

    Args:
        f: A function, class, or method.

    Returns:
        The fully qualified name of the module in which f is defined. If f is
        defined in the __main__ module, then the name of the file containing f
        (without its ".py" extension) is returned as the fully qualified module
        name. If f is defined in a Google Colab notebook, then "google_colab" is
        returned as the fully qualified module name.

    Raises:
        RuntimeError if f is defined in the __main__ module, the current
        environment is not Google Colab, and the name of the file
        containing f does not end with ".py".
    """
    qualname = f.__module__
    if qualname == "__main__":
        # Special case for Google Colab.
        try:
            import google.colab  # pylint: disable=import-outside-toplevel,unused-import
            return "google_colab"
        except ImportError:
            pass

        qualname, ext = os.path.splitext(
            os.path.basename(f.__globals__["__file__"]))
        if ext != ".py":
            raise RuntimeError(
                f"f ({f.__qualname__}) is defined in the __main__ module but "
                f"is contained in a file ({f.__globals__['__file__']}) that "
                "does not have extension '.py'.")
    return qualname
